@@ -29,6 +29,7 @@ enum {
   DOWNLOAD_REMOVED,
 
   ESTIMATED_PROGRESS_CHANGED,
+  SHOW_DOWNLOADS,
 
   LAST_SIGNAL
 };
@@ -131,6 +132,13 @@ ephy_downloads_manager_class_init (EphyDownloadsManagerClass *klass)
                   G_SIGNAL_RUN_LAST,
                   0, NULL, NULL, NULL,
                   G_TYPE_NONE, 0);
+
+  signals[SHOW_DOWNLOADS] =
+    g_signal_new ("show-downloads",
+                  EPHY_TYPE_DOWNLOADS_MANAGER,
+                  G_SIGNAL_RUN_LAST,
+                  0, NULL, NULL, NULL,
+                  G_TYPE_NONE, 0);
 }
 
 static void
@@ -159,24 +167,20 @@ download_estimated_progress_changed_cb (EphyDownloadsManager *manager)
   g_signal_emit (manager, signals[ESTIMATED_PROGRESS_CHANGED], 0);
 }
 
-static void
-download_created_destination_cb (EphyDownloadsManager *manager)
-{
-  ephy_downloads_manager_acquire_session_inhibitor (manager);
-}
-
 void
 ephy_downloads_manager_add_download (EphyDownloadsManager *manager,
                                      EphyDownload         *download)
 {
   WebKitDownload *wk_download;
 
-  g_return_if_fail (EPHY_IS_DOWNLOADS_MANAGER (manager));
-  g_return_if_fail (EPHY_IS_DOWNLOAD (download));
+  g_assert (EPHY_IS_DOWNLOADS_MANAGER (manager));
+  g_assert (EPHY_IS_DOWNLOAD (download));
 
   if (g_list_find (manager->downloads, download))
     return;
 
+  ephy_downloads_manager_acquire_session_inhibitor (manager);
+
   manager->downloads = g_list_prepend (manager->downloads, g_object_ref (download));
   g_signal_connect (download, "completed",
                     G_CALLBACK (download_completed_cb),
@@ -189,9 +193,6 @@ ephy_downloads_manager_add_download (EphyDownloadsManager *manager,
   g_signal_connect_swapped (wk_download, "notify::estimated-progress",
                             G_CALLBACK (download_estimated_progress_changed_cb),
                             manager);
-  g_signal_connect_swapped (wk_download, "created-destination",
-                            G_CALLBACK (download_created_destination_cb),
-                            manager);
   g_signal_emit (manager, signals[DOWNLOAD_ADDED], 0, download);
   g_signal_emit (manager, signals[ESTIMATED_PROGRESS_CHANGED], 0);
 }
@@ -202,8 +203,8 @@ ephy_downloads_manager_remove_download (EphyDownloadsManager *manager,
 {
   GList *download_link;
 
-  g_return_if_fail (EPHY_IS_DOWNLOADS_MANAGER (manager));
-  g_return_if_fail (EPHY_IS_DOWNLOAD (download));
+  g_assert (EPHY_IS_DOWNLOADS_MANAGER (manager));
+  g_assert (EPHY_IS_DOWNLOAD (download));
 
   download_link = g_list_find (manager->downloads, download);
   if (!download_link)
@@ -219,7 +220,7 @@ ephy_downloads_manager_has_active_downloads (EphyDownloadsManager *manager)
 {
   GList *l;
 
-  g_return_val_if_fail (EPHY_IS_DOWNLOADS_MANAGER (manager), FALSE);
+  g_assert (EPHY_IS_DOWNLOADS_MANAGER (manager));
 
   for (l = manager->downloads; l; l = g_list_next (l)) {
     EphyDownload *download = EPHY_DOWNLOAD (l->data);
@@ -234,7 +235,7 @@ ephy_downloads_manager_has_active_downloads (EphyDownloadsManager *manager)
 GList *
 ephy_downloads_manager_get_downloads (EphyDownloadsManager *manager)
 {
-  g_return_val_if_fail (EPHY_IS_DOWNLOADS_MANAGER (manager), NULL);
+  g_assert (EPHY_IS_DOWNLOADS_MANAGER (manager));
 
   return manager->downloads;
 }
@@ -246,7 +247,7 @@ ephy_downloads_manager_get_estimated_progress (EphyDownloadsManager *manager)
   guint n_active = 0;
   gdouble progress = 0;
 
-  g_return_val_if_fail (EPHY_IS_DOWNLOADS_MANAGER (manager), 0);
+  g_assert (EPHY_IS_DOWNLOADS_MANAGER (manager));
 
   for (l = manager->downloads; l; l = g_list_next (l)) {
     EphyDownload *download = EPHY_DOWNLOAD (l->data);
@@ -260,3 +261,19 @@ ephy_downloads_manager_get_estimated_progress (EphyDownloadsManager *manager)
 
   return n_active > 0 ? progress / n_active : 1;
 }
+
+EphyDownload *
+ephy_downloads_manager_find_download_by_id (EphyDownloadsManager *manager,
+                                            guint64               id)
+{
+  g_assert (EPHY_IS_DOWNLOADS_MANAGER (manager));
+
+  for (GList *l = manager->downloads; l; l = g_list_next (l)) {
+    EphyDownload *download = EPHY_DOWNLOAD (l->data);
+
+    if (ephy_download_get_uid (download) == id)
+      return download;
+  }
+
+  return NULL;
+}
