@@ -255,12 +255,10 @@ webkit_pref_callback_user_javascript (GSettings  *settings,
                      (GAsyncReadyCallback)user_javascript_read_cb, NULL);
 
   user_javascript_monitor = g_file_monitor_file (file, G_FILE_MONITOR_NONE, NULL, &error);
-  if (!user_javascript_monitor) {
+  if (!user_javascript_monitor)
     g_warning ("Could not create a file monitor for %s: %s\n", g_file_get_uri (file), error->message);
-    g_error_free (error);
-  } else {
+  else
     g_signal_connect (user_javascript_monitor, "changed", G_CALLBACK (user_javascript_file_changed), NULL);
-  }
 }
 
 static void
@@ -567,6 +565,8 @@ ephy_embed_prefs_init (gpointer user_data)
                                                        "enable-developer-extras", TRUE,
                                                        "enable-dns-prefetching", TRUE,
                                                        "enable-plugins", FALSE,
+                                                       "enable-media-stream", TRUE,
+                                                       "enable-smooth-scrolling", TRUE,
                                                        "javascript-can-open-windows-automatically", TRUE,
                                                        NULL);
 
@@ -599,10 +599,6 @@ ephy_embed_prefs_init (gpointer user_data)
                    EPHY_PREFS_WEB_DEFAULT_ENCODING,
                    webkit_settings, DEFAULT_ENCODING_SETTING,
                    G_SETTINGS_BIND_GET);
-  g_settings_bind (EPHY_SETTINGS_WEB,
-                   EPHY_PREFS_WEB_ENABLE_SMOOTH_SCROLLING,
-                   webkit_settings, "enable-smooth-scrolling",
-                   G_SETTINGS_BIND_GET);
   g_settings_bind (EPHY_SETTINGS_WEB,
                    EPHY_PREFS_WEB_ENABLE_SITE_SPECIFIC_QUIRKS,
                    webkit_settings, "enable-site-specific-quirks",
