@@ -36,18 +36,14 @@ webkit_web_extension_initialize_with_user_data (WebKitWebExtension *webkit_exten
                                                 GVariant           *user_data)
 {
   const char *guid;
-  const char *server_address;
   const char *profile_dir;
+  const char *webextension_translations;
   gboolean private_profile;
   gboolean should_remember_passwords;
+  gboolean is_webextension;
   g_autoptr (GError) error = NULL;
 
-  g_variant_get (user_data, "(&sm&sm&sbb)", &guid, &server_address, &profile_dir, &should_remember_passwords, &private_profile);
-
-  if (!server_address) {
-    g_warning ("UI process did not start D-Bus server, giving up.");
-    return;
-  }
+  g_variant_get (user_data, "(&sm&sbbb&s)", &guid, &profile_dir, &should_remember_passwords, &private_profile, &is_webextension, &webextension_translations);
 
   if (!ephy_file_helpers_init (profile_dir, 0, &error))
     g_warning ("Failed to initialize file helpers: %s", error->message);
@@ -59,10 +55,12 @@ webkit_web_extension_initialize_with_user_data (WebKitWebExtension *webkit_exten
 
   extension = ephy_web_process_extension_get ();
 
+  if (is_webextension)
+    return;
+
   ephy_web_process_extension_initialize (extension,
                                          webkit_extension,
                                          guid,
-                                         server_address,
                                          should_remember_passwords,
                                          private_profile);
 }
@@ -70,9 +68,7 @@ webkit_web_extension_initialize_with_user_data (WebKitWebExtension *webkit_exten
 static void __attribute__((destructor))
 ephy_web_process_extension_shutdown (void)
 {
-  if (extension)
-    g_object_unref (extension);
-
+  g_clear_object (&extension);
   ephy_settings_shutdown ();
   ephy_file_helpers_shutdown ();
 }
