@@ -29,6 +29,7 @@
 #include "ephy-settings.h"
 #include "ephy-uri-helpers.h"
 #include "ephy-web-overview-model.h"
+#include "ephy-webextension-common.h"
 
 #include <gio/gio.h>
 #include <glib/gi18n.h>
@@ -52,6 +53,7 @@ struct _EphyWebProcessExtension {
   EphyPermissionsManager *permissions_manager;
 
   WebKitScriptWorld *script_world;
+  GHashTable *content_script_worlds;
 
   gboolean should_remember_passwords;
   gboolean is_private_profile;
@@ -62,12 +64,6 @@ struct _EphyWebProcessExtension {
 
 G_DEFINE_TYPE (EphyWebProcessExtension, ephy_web_process_extension, G_TYPE_OBJECT)
 
-GHashTable *
-ephy_web_process_extension_get_translations (EphyWebProcessExtension *extension)
-{
-  return extension->translation_table;
-}
-
 static void
 web_page_will_submit_form (WebKitWebPage            *web_page,
                            WebKitDOMHTMLFormElement *dom_form,
@@ -198,6 +194,8 @@ web_page_context_menu (WebKitWebPage          *web_page,
 {
   EphyWebProcessExtension *extension;
   g_autofree char *string = NULL;
+  gboolean is_editable;
+  gboolean is_password;
   GVariantBuilder builder;
   WebKitFrame *frame;
   g_autoptr (JSCContext) js_context = NULL;
@@ -213,28 +211,106 @@ web_page_context_menu (WebKitWebPage          *web_page,
   js_value = jsc_context_evaluate (js_context, "window.getSelection().toString();", -1);
   if (!jsc_value_is_null (js_value) && !jsc_value_is_undefined (js_value))
     string = jsc_value_to_string (js_value);
+  g_object_unref (js_value);
 
-  if (!string || *string == '\0')
-    return FALSE;
+  js_value = jsc_context_evaluate (js_context, "contextMenuElementIsEditable;", -1);
+  is_editable = jsc_value_to_boolean (js_value);
+  g_object_unref (js_value);
+
+  js_value = jsc_context_evaluate (js_context, "contextMenuElementIsPassword;", -1);
+  is_password = jsc_value_to_boolean (js_value);
 
   g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
-  g_variant_builder_add (&builder, "{sv}", "SelectedText", g_variant_new_string (g_strstrip (string)));
+  g_variant_builder_add (&builder, "{sv}", "SelectedText", g_variant_new_string (string ? g_strstrip (string) : ""));
+  g_variant_builder_add (&builder, "{sv}", "IsEditable", g_variant_new_boolean (is_editable));
+  g_variant_builder_add (&builder, "{sv}", "IsPassword", g_variant_new_boolean (is_password));
   webkit_context_menu_set_user_data (context_menu,
                                      g_variant_builder_end (&builder));
 
   return TRUE;
 }
 
+static void
+content_script_window_object_cleared_cb (WebKitScriptWorld *world,
+                                         WebKitWebPage     *page,
+                                         WebKitFrame       *frame,
+                                         gpointer           user_data)
+{
+  EphyWebProcessExtension *extension = user_data;
+  g_autoptr (JSCContext) js_context = NULL;
+  g_autoptr (JSCValue) js_browser = NULL;
+  g_autoptr (JSCValue) result = NULL;
+  g_autoptr (GBytes) bytes = NULL;
+  JsonObject *translations;
+  const char *guid;
+  const char *data;
+  gsize data_size;
+
+  guid = webkit_script_world_get_name (world);
+  js_context = webkit_frame_get_js_context_for_script_world (frame, world);
+  translations = g_hash_table_lookup (extension->translation_table, guid);
+
+  js_browser = jsc_context_get_value (js_context, "browser");
+  g_assert (!jsc_value_is_object (js_browser));
+
+  bytes = g_resources_lookup_data ("/org/gnome/epiphany-web-extension/js/webextensions-common.js", G_RESOURCE_LOOKUP_FLAGS_NONE, NULL);
+  data = g_bytes_get_data (bytes, &data_size);
+  result = jsc_context_evaluate_with_source_uri (js_context, data, data_size, "resource:///org/gnome/epiphany-web-extension/js/webextensions-common.js", 1);
+  g_clear_object (&result);
+
+  ephy_webextension_install_common_apis (page,
+                                         frame,
+                                         js_context,
+                                         guid,
+                                         translations);
+}
+
+static void
+create_content_script_world (EphyWebProcessExtension *extension,
+                             const char              *guid)
+{
+  WebKitScriptWorld *world = webkit_script_world_new_with_name (guid);
+
+  g_hash_table_insert (extension->content_script_worlds, g_strdup (guid), world);
+
+  g_signal_connect (world,
+                    "window-object-cleared",
+                    G_CALLBACK (content_script_window_object_cleared_cb),
+                    extension);
+}
+
+static gboolean
+web_page_received_message (WebKitWebPage     *web_page,
+                           WebKitUserMessage *message,
+                           gpointer           user_data)
+{
+  EphyWebProcessExtension *extension = user_data;
+  const char *name = webkit_user_message_get_name (message);
+
+  if (g_strcmp0 (name, "WebExtension.Initialize") == 0) {
+    GVariant *parameters;
+    const char *guid;
+
+    parameters = webkit_user_message_get_parameters (message);
+    if (!parameters)
+      return FALSE;
+
+    g_variant_get (parameters, "&s", &guid);
+    create_content_script_world (extension, guid);
+  } else {
+    g_warning ("Unhandled page message: %s", name);
+    return FALSE;
+  }
+
+  return TRUE;
+}
+
 static void
 ephy_web_process_extension_page_created_cb (EphyWebProcessExtension *extension,
                                             WebKitWebPage           *web_page)
 {
   g_autoptr (JSCContext) js_context = NULL;
 
-  /* Enforce the creation of the script world global context in the main frame */
-  js_context = webkit_frame_get_js_context_for_script_world (webkit_web_page_get_main_frame (web_page), extension->script_world);
-  (void)js_context;
-
   g_signal_connect (web_page, "context-menu",
                     G_CALLBACK (web_page_context_menu),
                     extension);
@@ -244,6 +320,37 @@ ephy_web_process_extension_page_created_cb (EphyWebProcessExtension *extension,
   g_signal_connect (web_page, "form-controls-associated-for-frame",
                     G_CALLBACK (web_page_form_controls_associated),
                     extension);
+  g_signal_connect (web_page, "user-message-received",
+                    G_CALLBACK (web_page_received_message),
+                    extension);
+}
+
+static void
+ephy_web_process_extension_add_translations (GHashTable *translation_table,
+                                             const char *guid,
+                                             const char *data)
+{
+  g_autoptr (JsonParser) parser = NULL;
+  JsonNode *root;
+  JsonObject *root_object;
+  g_autoptr (GError) error = NULL;
+
+  g_hash_table_remove (translation_table, guid);
+
+  if (!data || !*data)
+    return;
+
+  parser = json_parser_new ();
+  if (json_parser_load_from_data (parser, data, -1, &error)) {
+    root = json_parser_get_root (parser);
+    g_assert (root);
+    root_object = json_node_get_object (root);
+    g_assert (root_object);
+
+    g_hash_table_insert (translation_table, g_strdup (guid), json_object_ref (root_object));
+  } else {
+    g_warning ("Could not read translation json data: %s. '%s'", error->message, data);
+  }
 }
 
 static void
@@ -334,18 +441,17 @@ ephy_web_process_extension_user_message_received_cb (EphyWebProcessExtension *ex
       return;
 
     g_variant_get (parameters, "b", &extension->should_remember_passwords);
-  } else if (g_strcmp0 (name, "WebExtension.Add") == 0) {
+  } else if (g_strcmp0 (name, "WebExtension.UpdateTranslations") == 0) {
     GVariant *parameters;
-    const char *name;
+    const char *guid;
     const char *data;
-    guint64 length;
 
     parameters = webkit_user_message_get_parameters (message);
     if (!parameters)
       return;
 
-    g_variant_get (parameters, "(&s&st)", &name, &data, &length);
-    webextensions_add_translation (extension, name, data, length);
+    g_variant_get (parameters, "(&s&s)", &guid, &data);
+    ephy_web_process_extension_add_translations (extension->translation_table, guid, data);
   }
 }
 
@@ -397,6 +503,7 @@ ephy_web_process_extension_dispose (GObject *object)
   }
 
   g_clear_pointer (&extension->translation_table, g_hash_table_destroy);
+  g_clear_pointer (&extension->content_script_worlds, g_hash_table_destroy);
 
   G_OBJECT_CLASS (ephy_web_process_extension_parent_class)->dispose (object);
 }
@@ -667,27 +774,9 @@ window_object_cleared_cb (WebKitScriptWorld       *world,
   js_context = webkit_frame_get_js_context_for_script_world (frame, world);
   jsc_context_push_exception_handler (js_context, (JSCExceptionHandler)js_exception_handler, NULL, NULL);
 
-  /* If we are using the default script world, then we are a WebExtension. We
-   * must not register any internal Epiphany APIs, since they must never be
-   * accessible in the default script world. We don't want them exposed to the
-   * web or to WebExtensions. If we were to improperly allow access to our
-   * internal APIs, then malicious web content could do nasty things like
-   * iterate through passwords stored in EphyPasswordsManager, for example.
-   *
-   * And if we are not using the default script world, then we are not a
-   * WebExtension. There is no point in registering WebExtension APIs, because
-   * WebExtensions only have access to what is in the default script world
-   * anyway.
-   *
-   * FIXME: let's try to make this less confusing:
-   *
-   * https://gitlab.gnome.org/GNOME/epiphany/-/issues/1448
-   * https://gitlab.gnome.org/GNOME/epiphany/-/issues/1449
-   */
-  if (extension->script_world == webkit_script_world_get_default ()) {
-    set_up_webextensions (extension, page, js_context);
-    return;
-  }
+  result = jsc_context_get_value (js_context, "Ephy");
+  g_assert (jsc_value_is_undefined (result));
+  g_clear_object (&result);
 
   bytes = g_resources_lookup_data ("/org/gnome/epiphany-web-process-extension/js/ephy.js", G_RESOURCE_LOOKUP_FLAGS_NONE, NULL);
   data = g_bytes_get_data (bytes, &data_size);
@@ -815,11 +904,9 @@ ephy_web_process_extension_initialize (EphyWebProcessExtension *extension,
 
   extension->initialized = TRUE;
 
-  /* Note: An empty guid is used ONLY for WebExtensions which do have an own initialization function */
-  if (strlen (guid) > 0)
-    extension->script_world = webkit_script_world_new_with_name (guid);
-  else
-    extension->script_world = webkit_script_world_get_default ();
+  g_assert (guid && *guid);
+
+  extension->script_world = webkit_script_world_new_with_name (guid);
 
   g_signal_connect (extension->script_world,
                     "window-object-cleared",
@@ -843,5 +930,9 @@ ephy_web_process_extension_initialize (EphyWebProcessExtension *extension,
   extension->frames_map = g_hash_table_new_full (g_int64_hash, g_int64_equal,
                                                  g_free, NULL);
 
-  extension->translation_table = g_hash_table_new (g_str_hash, NULL);
+  extension->translation_table = g_hash_table_new_full (g_str_hash, g_str_equal,
+                                                        g_free, (GDestroyNotify)json_object_unref);
+
+  extension->content_script_worlds = g_hash_table_new_full (g_str_hash, g_str_equal,
+                                                            g_free, g_object_unref);
 }
