@@ -114,11 +114,10 @@ ephy_string_shorten (char  *str,
   /* create string */
   bytes = GPOINTER_TO_UINT (g_utf8_offset_to_pointer (str, target_length - 1) - str);
 
-  /* +1 for ellipsis, +1 for trailing NUL */
-  new_str = g_new (gchar, bytes + 1 + 1);
+  new_str = g_new (gchar, bytes + strlen ("…") + 1);
 
   strncpy (new_str, str, bytes);
-  strcat (new_str, "…");
+  strncpy (new_str + bytes, "…", strlen ("…") + 1);
 
   g_free (str);
 
@@ -192,7 +191,8 @@ ephy_string_get_host_name (const char *url)
     g_free (effective_url);
   }
 
-  if (uri == NULL) return NULL;
+  if (uri == NULL)
+    return NULL;
 
   return g_strdup (g_uri_get_host (uri));
 }
@@ -244,23 +244,13 @@ ephy_string_find_and_replace (const char *haystack,
                               const char *to_repl)
 {
   GString *str;
-  const char *tmp;
-  gsize to_find_len;
-  gsize pos;
 
   g_assert (haystack);
   g_assert (to_find);
   g_assert (to_repl);
 
   str = g_string_new (haystack);
-  to_find_len = strlen (to_find);
-
-  while ((tmp = strstr (str->str, to_find)) != NULL) {
-    pos = tmp - str->str;
-    g_string_erase (str, pos, to_find_len);
-    g_string_insert (str, pos, to_repl);
-  }
-
+  g_string_replace (str, to_find, to_repl, 0);
   return g_string_free (str, FALSE);
 }
 
