@@ -21,6 +21,8 @@
 
 #pragma once
 
+#include "ephy-bookmark-states.h"
+
 #include <gtk/gtk.h>
 
 G_BEGIN_DECLS
@@ -32,9 +34,17 @@ G_DECLARE_FINAL_TYPE (EphyActionBarEnd, ephy_action_bar_end, EPHY, ACTION_BAR_EN
 EphyActionBarEnd *ephy_action_bar_end_new                       (void);
 void              ephy_action_bar_end_set_show_bookmarks_button (EphyActionBarEnd *action_bar_end,
                                                                  gboolean          show);
+void              ephy_action_bar_end_show_downloads            (EphyActionBarEnd *action_bar_end);
+void              ephy_action_bar_end_show_bookmarks            (EphyActionBarEnd *action_bar_end);
 GtkWidget        *ephy_action_bar_end_get_downloads_revealer    (EphyActionBarEnd *action_bar_end);
 
 void              ephy_action_bar_end_add_browser_action        (EphyActionBarEnd *action_bar_end,
                                                                  GtkWidget        *action);
 
+void              ephy_action_bar_end_set_show_bookmark_button (EphyActionBarEnd *action_bar_end,
+                                                                gboolean          show);
+
+void              ephy_action_bar_end_set_bookmark_icon_state (EphyActionBarEnd       *action_bar_end,
+                                                               EphyBookmarkIconState  state);
+
 G_END_DECLS
