@@ -492,7 +492,7 @@ sync_setup_firefox_iframe (EphyFirefoxSyncDialog *sync_dialog)
                        GTK_WIDGET (sync_dialog->fxa_web_view));
     gtk_box_pack_start (GTK_BOX (sync_dialog->sync_firefox_iframe_box),
                         frame,
-                        TRUE, TRUE, 0);
+                        FALSE, TRUE, 0);
 
     g_object_unref (sync_context);
   }
@@ -538,7 +538,7 @@ on_sync_synced_tabs_button_clicked (GtkWidget             *button,
   synced_tabs_dialog = synced_tabs_dialog_new (manager);
   gtk_window_set_transient_for (GTK_WINDOW (synced_tabs_dialog), GTK_WINDOW (sync_dialog));
   gtk_window_set_modal (GTK_WINDOW (synced_tabs_dialog), TRUE);
-  gtk_window_present_with_time (GTK_WINDOW (synced_tabs_dialog), gtk_get_current_event_time ());
+  gtk_window_present (GTK_WINDOW (synced_tabs_dialog));
 }
 
 static void
@@ -699,7 +699,10 @@ create_sync_frequency_minutes_model ()
 static gchar *
 get_sync_frequency_minutes_name (HdyValueObject *value)
 {
-  return g_strdup_printf ("%u min", g_value_get_uint (hdy_value_object_get_value (value)));
+  guint n_minutes = g_value_get_uint (hdy_value_object_get_value (value));
+  const char *minutes_text = ngettext ("%u min", "%u mins", n_minutes);
+
+  return g_strdup_printf (minutes_text, n_minutes);
 }
 
 void
