@@ -26,17 +26,16 @@
 #include "ephy-file-chooser.h"
 #include "ephy-file-helpers.h"
 #include "ephy-flatpak-utils.h"
+#include "ephy-lang-row.h"
 #include "ephy-langs.h"
 #include "ephy-settings.h"
-#include "ephy-search-engine-dialog.h"
+#include "ephy-search-engine-listbox.h"
 #include "ephy-web-app-utils.h"
 #include "webapp-additional-urls-dialog.h"
 
 #include "gnome-languages.h"
 #include <glib/gi18n.h>
 
-#define DOWNLOAD_BUTTON_WIDTH   8
-
 enum {
   COL_LANG_NAME,
   COL_LANG_CODE
@@ -50,8 +49,11 @@ struct _PrefsGeneralPage {
   guint webapp_save_id;
   GtkWidget *webapp_box;
   GtkWidget *webapp_icon;
+  GtkWidget *webapp_icon_row;
   GtkWidget *webapp_url;
+  GtkWidget *webapp_url_row;
   GtkWidget *webapp_title;
+  GtkWidget *webapp_title_row;
 
   /* Web Content */
   GtkWidget *adblock_allow_switch;
@@ -68,9 +70,10 @@ struct _PrefsGeneralPage {
   GtkWidget *download_box;
   GtkWidget *ask_on_download_switch;
   GtkWidget *download_folder_row;
+  GtkWidget *download_folder_label;
 
   /* Search Engines */
-  GtkWidget *search_box;
+  GtkWidget *search_engine_group;
 
   /* Session */
   GtkWidget *session_box;
@@ -80,7 +83,6 @@ struct _PrefsGeneralPage {
 
   /* Browsing */
   GtkWidget *browsing_box;
-  GtkWidget *enable_smooth_scrolling_switch;
   GtkWidget *enable_mouse_gesture_switch;
   GtkWidget *enable_switch_to_new_tab;
 
@@ -113,21 +115,24 @@ prefs_general_page_finalize (GObject *object)
 }
 
 static int
-get_list_box_length (GtkWidget *list_box)
+get_list_box_length (GtkListBox *list_box)
 {
-  GList *children = gtk_container_get_children (GTK_CONTAINER (list_box));
+  int i = -1;
 
-  return g_list_length (children);
+  while (gtk_list_box_get_row_at_index (list_box, ++i));
+
+  return i;
 }
 
 static void
 language_editor_update_pref (PrefsGeneralPage *general_page)
 {
   GVariantBuilder builder;
-  GtkListBoxRow *row;
-  int index = 0;
+  GtkListBox *lang_listbox = GTK_LIST_BOX (general_page->lang_listbox);
+  int len = get_list_box_length (lang_listbox);
+  int index;
 
-  if (get_list_box_length (general_page->lang_listbox) <= 1) {
+  if (get_list_box_length (lang_listbox) <= 1) {
     g_settings_set (EPHY_SETTINGS_WEB,
                     EPHY_PREFS_WEB_LANGUAGE,
                     "as", NULL);
@@ -136,10 +141,10 @@ language_editor_update_pref (PrefsGeneralPage *general_page)
 
   g_variant_builder_init (&builder, G_VARIANT_TYPE_STRING_ARRAY);
 
-  while ((row = gtk_list_box_get_row_at_index (GTK_LIST_BOX (general_page->lang_listbox), index++))) {
-    char *code;
+  for (index = 0; index < len - 1; index++) {
+    GtkListBoxRow *row = gtk_list_box_get_row_at_index (lang_listbox, index);
+    const char *code = ephy_lang_row_get_code (EPHY_LANG_ROW (row));
 
-    code = g_object_get_data (G_OBJECT (row), "code");
     if (code)
       g_variant_builder_add (&builder, "s", code);
   }
@@ -159,78 +164,46 @@ drag_data_received (GtkWidget        *widget,
                     guint32           time,
                     gpointer          data)
 {
-  GtkWidget *row_before;
-  GtkWidget *row_after;
-  GtkWidget *row;
-  GtkWidget *source;
   PrefsGeneralPage *general_page = data;
-  int len;
-  int pos;
+  GtkListBox *lang_listbox = GTK_LIST_BOX (general_page->lang_listbox);
+  GtkWidget *hovered_row;
+  GtkWidget *dragged_row;
+  int length;
+  int hovered_row_pos;
+  int dragged_row_pos;
+  int insert_pos;
 
-  row_before = GTK_WIDGET (g_object_get_data (G_OBJECT (widget), "row-before"));
-  row_after = GTK_WIDGET (g_object_get_data (G_OBJECT (widget), "row-after"));
+  hovered_row = GTK_WIDGET (gtk_list_box_get_row_at_y (lang_listbox, y));
+  dragged_row = (gpointer) * (gpointer *)gtk_selection_data_get_data (selection_data);
 
-  g_object_set_data (G_OBJECT (widget), "row-before", NULL);
-  g_object_set_data (G_OBJECT (widget), "row-after", NULL);
+  length = get_list_box_length (GTK_LIST_BOX (widget));
+  hovered_row_pos = gtk_list_box_row_get_index (GTK_LIST_BOX_ROW (hovered_row));
+  dragged_row_pos = gtk_list_box_row_get_index (GTK_LIST_BOX_ROW (dragged_row));
 
-  if (row_before)
-    gtk_style_context_remove_class (gtk_widget_get_style_context (row_before), "drag-hover-bottom");
-  if (row_after)
-    gtk_style_context_remove_class (gtk_widget_get_style_context (row_after), "drag-hover-top");
-
-  row = (gpointer) * (gpointer *)gtk_selection_data_get_data (selection_data);
-  source = gtk_widget_get_ancestor (row, GTK_TYPE_LIST_BOX_ROW);
-
-  if (source == row_after)
+  /* If the drag ended while hovering over the dragged row then we return */
+  if (hovered_row == dragged_row)
     return;
 
-  len = get_list_box_length (general_page->lang_listbox);
-  g_object_ref (source);
-  gtk_container_remove (GTK_CONTAINER (gtk_widget_get_parent (source)), source);
+  /* If the drag ended while hovering over the "Add Language" row then we return */
+  if (hovered_row_pos == length - 1)
+    return;
 
-  if (row_after)
-    pos = gtk_list_box_row_get_index (GTK_LIST_BOX_ROW (row_after));
-  else
-    pos = gtk_list_box_row_get_index (GTK_LIST_BOX_ROW (row_before)) + 1;
+  g_object_ref (dragged_row);
+  gtk_container_remove (GTK_CONTAINER (lang_listbox), dragged_row);
 
-  if (pos + 1 == len)
-    pos--;
-
-  gtk_list_box_insert (GTK_LIST_BOX (widget), source, pos);
-  g_object_unref (source);
-
-  language_editor_update_pref (general_page);
-}
-
-static GtkListBoxRow *
-get_row_before (GtkListBox    *list,
-                GtkListBoxRow *row)
-{
-  int pos = gtk_list_box_row_get_index (row);
-  return gtk_list_box_get_row_at_index (list, pos - 1);
-}
-
-static GtkListBoxRow *
-get_row_after (GtkListBox    *list,
-               GtkListBoxRow *row)
-{
-  int pos = gtk_list_box_row_get_index (row);
-  return gtk_list_box_get_row_at_index (list, pos + 1);
-}
-
-static GtkListBoxRow *
-get_last_row (GtkListBox *list)
-{
-  int i;
-
-  for (i = 0;; i++) {
-    GtkListBoxRow *tmp;
-    tmp = gtk_list_box_get_row_at_index (list, i);
-    if (tmp == NULL)
-      break;
+  /* Calculation logic for insert_pos */
+  if (y < 20) {
+    insert_pos = 0;
+  } else if (dragged_row_pos < hovered_row_pos) {
+    insert_pos = hovered_row_pos;
+  } else { /* dragged_row_pos > hovered_row_pos is true here */
+    insert_pos = hovered_row_pos + 1;
   }
 
-  return i > 0 ? gtk_list_box_get_row_at_index (list, i - 1) : NULL;
+  gtk_list_box_insert (lang_listbox, dragged_row, insert_pos);
+  g_object_unref (dragged_row);
+
+  language_editor_update_pref (general_page);
 }
 
 static gboolean
@@ -240,77 +213,57 @@ drag_motion (GtkWidget      *widget,
              int             y,
              guint           time)
 {
-  GtkAllocation alloc;
-  GtkWidget *row;
-  int hover_row_y;
-  int hover_row_height;
-  GtkWidget *drag_row;
-  GtkWidget *row_before;
-  GtkWidget *row_after;
+  GtkListBoxRow *first_lang_row = gtk_list_box_get_row_at_index (GTK_LIST_BOX (widget), 0);
 
-  row = GTK_WIDGET (gtk_list_box_get_row_at_y (GTK_LIST_BOX (widget), y));
+  GtkWidget *dragged_row;
+  GtkWidget *hovered_row;
+  GtkWidget *expanded_dnd_revealer;
+  GtkWidget *hovered_row_dnd_revealer = NULL;
 
-  drag_row = GTK_WIDGET (g_object_get_data (G_OBJECT (widget), "drag-row"));
-  row_before = GTK_WIDGET (g_object_get_data (G_OBJECT (widget), "row-before"));
-  row_after = GTK_WIDGET (g_object_get_data (G_OBJECT (widget), "row-after"));
+  dragged_row = g_object_get_data (G_OBJECT (widget), "dragged-row");
+  hovered_row = GTK_WIDGET (gtk_list_box_get_row_at_y (GTK_LIST_BOX (widget), y));
+  expanded_dnd_revealer = g_object_get_data (G_OBJECT (widget), "dnd-expanded-revealer");
 
-  gtk_style_context_remove_class (gtk_widget_get_style_context (drag_row), "drag-hover");
-  if (row_before)
-    gtk_style_context_remove_class (gtk_widget_get_style_context (row_before), "drag-hover-bottom");
-  if (row_after)
-    gtk_style_context_remove_class (gtk_widget_get_style_context (row_after), "drag-hover-top");
+  if (EPHY_IS_LANG_ROW (hovered_row))
+    hovered_row_dnd_revealer = ephy_lang_row_get_dnd_bottom_revealer (EPHY_LANG_ROW (hovered_row));
 
-  if (row) {
-    gtk_widget_get_allocation (row, &alloc);
-    hover_row_y = alloc.y;
-    hover_row_height = alloc.height;
+  /* Edge case: If the user is hovering with the dragged row close to the
+   * top of the list, then we reveal the top empty row of EphyLangRow */
+  if (y < 20 && dragged_row != GTK_WIDGET (first_lang_row)) {
+    GtkWidget *top_revealer = ephy_lang_row_get_dnd_top_revealer (EPHY_LANG_ROW (hovered_row));
 
-    if (y < hover_row_y + hover_row_height / 2) {
-      row_after = row;
-      row_before = GTK_WIDGET (get_row_before (GTK_LIST_BOX (widget), GTK_LIST_BOX_ROW (row)));
-    } else {
-      row_before = row;
-      row_after = GTK_WIDGET (get_row_after (GTK_LIST_BOX (widget), GTK_LIST_BOX_ROW (row)));
-    }
-  } else {
-    row_before = GTK_WIDGET (get_last_row (GTK_LIST_BOX (widget)));
-    row_after = NULL;
+    if (expanded_dnd_revealer)
+      gtk_revealer_set_reveal_child (GTK_REVEALER (expanded_dnd_revealer), FALSE);
+
+    gtk_revealer_set_reveal_child (GTK_REVEALER (top_revealer), TRUE);
+    g_object_set_data (G_OBJECT (widget), "dnd-expanded-revealer", top_revealer);
+
+    return TRUE;
   }
 
-  g_object_set_data (G_OBJECT (widget), "row-before", row_before);
-  g_object_set_data (G_OBJECT (widget), "row-after", row_after);
+  if (dragged_row == hovered_row) {
+    if (expanded_dnd_revealer)
+      gtk_revealer_set_reveal_child (GTK_REVEALER (expanded_dnd_revealer), FALSE);
 
-  if (drag_row == row_before || drag_row == row_after) {
-    gtk_style_context_add_class (gtk_widget_get_style_context (drag_row), "drag-hover");
+    g_object_set_data (G_OBJECT (widget), "dnd-expanded-revealer", NULL);
+
+    return TRUE;
+  }
+
+  if (EPHY_IS_LANG_ROW (hovered_row)) {
+    if (expanded_dnd_revealer)
+      gtk_revealer_set_reveal_child (GTK_REVEALER (expanded_dnd_revealer), FALSE);
+
+    gtk_revealer_set_reveal_child (GTK_REVEALER (hovered_row_dnd_revealer), TRUE);
+    g_object_set_data (G_OBJECT (widget), "dnd-expanded-revealer", hovered_row_dnd_revealer);
+    return TRUE;
+  } else {
+    if (expanded_dnd_revealer)
+      gtk_revealer_set_reveal_child (GTK_REVEALER (expanded_dnd_revealer), FALSE);
+
+    g_object_set_data (G_OBJECT (widget), "dnd-expanded-revealer", NULL);
     return FALSE;
   }
-
-  if (row_before)
-    gtk_style_context_add_class (gtk_widget_get_style_context (row_before), "drag-hover-bottom");
-  if (row_after)
-    gtk_style_context_add_class (gtk_widget_get_style_context (row_after), "drag-hover-top");
-
-  return TRUE;
-}
-
-static void
-drag_leave (GtkWidget      *widget,
-            GdkDragContext *context,
-            guint           time)
-{
-  GtkWidget *drag_row;
-  GtkWidget *row_before;
-  GtkWidget *row_after;
-
-  drag_row = GTK_WIDGET (g_object_get_data (G_OBJECT (widget), "drag-row"));
-  row_before = GTK_WIDGET (g_object_get_data (G_OBJECT (widget), "row-before"));
-  row_after = GTK_WIDGET (g_object_get_data (G_OBJECT (widget), "row-after"));
-
-  gtk_style_context_remove_class (gtk_widget_get_style_context (drag_row), "drag-hover");
-  if (row_before)
-    gtk_style_context_remove_class (gtk_widget_get_style_context (row_before), "drag-hover-bottom");
-  if (row_after)
-    gtk_style_context_remove_class (gtk_widget_get_style_context (row_after), "drag-hover-top");
 }
 
 static GtkDialog *setup_add_language_dialog (PrefsGeneralPage *general_page);
@@ -335,7 +288,7 @@ language_editor_add_button_release_event (GtkWidget        *button,
       (gpointer *)add_lang_dialog);
   }
 
-  gtk_window_present_with_time (GTK_WINDOW (general_page->add_lang_dialog), gtk_get_current_event_time ());
+  gtk_window_present (GTK_WINDOW (general_page->add_lang_dialog));
 }
 
 static void
@@ -357,36 +310,33 @@ language_editor_add_function_buttons (PrefsGeneralPage *general_page)
 static void
 language_editor_update_state (PrefsGeneralPage *general_page)
 {
-  int length = get_list_box_length (general_page->lang_listbox);
+  GtkListBox *lang_listbox = GTK_LIST_BOX (general_page->lang_listbox);
+  int length = get_list_box_length (lang_listbox);
   int index;
 
+  /* If there's only one language row in the list we want to make its
+   * remove button insensitive */
   if (length == 2) {
-    GtkListBoxRow *row = gtk_list_box_get_row_at_index (GTK_LIST_BOX (general_page->lang_listbox), 0);
-    GtkWidget *action = g_object_get_data (G_OBJECT (row), "action");
+    GtkListBoxRow *row = gtk_list_box_get_row_at_index (lang_listbox, 0);
 
-    gtk_widget_set_sensitive (action, FALSE);
+    ephy_lang_row_set_delete_sensitive (EPHY_LANG_ROW (row), FALSE);
     return;
   }
 
   for (index = 0; index < length - 1; index++) {
-    GtkListBoxRow *row = gtk_list_box_get_row_at_index (GTK_LIST_BOX (general_page->lang_listbox), index);
-    GtkWidget *action = g_object_get_data (G_OBJECT (row), "action");
+    GtkListBoxRow *row = gtk_list_box_get_row_at_index (lang_listbox, index);
 
-    gtk_widget_set_sensitive (action, TRUE);
+    ephy_lang_row_set_delete_sensitive (EPHY_LANG_ROW (row), TRUE);
   }
 }
 
 static void
-language_editor_remove_button_clicked_cb (GtkWidget        *button,
+language_editor_delete_button_clicked_cb (EphyLangRow      *row,
                                           PrefsGeneralPage *general_page)
 {
-  GtkWidget *row = g_object_get_data (G_OBJECT (button), "row");
-
-  if (row) {
-    gtk_container_remove (GTK_CONTAINER (general_page->lang_listbox), row);
-    language_editor_update_pref (general_page);
-    language_editor_update_state (general_page);
-  }
+  gtk_container_remove (GTK_CONTAINER (general_page->lang_listbox), GTK_WIDGET (row));
+  language_editor_update_pref (general_page);
+  language_editor_update_state (general_page);
 }
 
 void
@@ -397,10 +347,12 @@ drag_data_get (GtkWidget        *widget,
                guint             time,
                gpointer          data)
 {
+  GtkWidget *row = gtk_widget_get_ancestor (widget, EPHY_TYPE_LANG_ROW);
+
   gtk_selection_data_set (selection_data,
-                          gdk_atom_intern_static_string ("GTK_LIST_BOX_ROW"),
+                          gdk_atom_intern_static_string ("EPHY_LANG_ROW"),
                           32,
-                          (const guchar *)&widget,
+                          (const guchar *)&row,
                           sizeof (gpointer));
 }
 
@@ -416,7 +368,7 @@ drag_begin (GtkWidget      *widget,
   int x, y;
   double sx, sy;
 
-  row = gtk_widget_get_ancestor (widget, GTK_TYPE_LIST_BOX_ROW);
+  row = gtk_widget_get_ancestor (widget, EPHY_TYPE_LANG_ROW);
   gtk_widget_get_allocation (row, &alloc);
   surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, alloc.width, alloc.height);
   cr = cairo_create (surface);
@@ -433,7 +385,7 @@ drag_begin (GtkWidget      *widget,
   cairo_destroy (cr);
   cairo_surface_destroy (surface);
 
-  g_object_set_data (G_OBJECT (gtk_widget_get_parent (row)), "drag-row", row);
+  g_object_set_data (G_OBJECT (gtk_widget_get_parent (row)), "dragged-row", row);
   gtk_style_context_add_class (gtk_widget_get_style_context (row), "drag-row");
 }
 
@@ -442,12 +394,21 @@ drag_end (GtkWidget      *widget,
           GdkDragContext *context,
           gpointer        data)
 {
-  GtkWidget *row;
+  GtkWidget *lang_row;
+  GtkWidget *lang_listbox;
+  GtkWidget *dnd_expanded_revealer;
 
-  row = gtk_widget_get_ancestor (widget, GTK_TYPE_LIST_BOX_ROW);
-  g_object_set_data (G_OBJECT (gtk_widget_get_parent (row)), "drag-row", NULL);
-  gtk_style_context_remove_class (gtk_widget_get_style_context (row), "drag-row");
-  gtk_style_context_remove_class (gtk_widget_get_style_context (row), "drag-hover");
+  lang_row = gtk_widget_get_ancestor (widget, EPHY_TYPE_LANG_ROW);
+  lang_listbox = gtk_widget_get_parent (lang_row);
+  dnd_expanded_revealer = g_object_get_data (G_OBJECT (lang_listbox), "dnd-expanded-revealer");
+
+  g_object_set_data (G_OBJECT (lang_listbox), "dragged-row", NULL);
+  gtk_style_context_remove_class (gtk_widget_get_style_context (lang_row), "drag-row");
+
+  if (dnd_expanded_revealer) {
+    gtk_revealer_set_reveal_child (GTK_REVEALER (dnd_expanded_revealer), FALSE);
+    g_object_set_data (G_OBJECT (lang_listbox), "dnd-expanded-revealer", NULL);
+  }
 }
 
 static GtkTargetEntry entries[] = {
@@ -459,53 +420,41 @@ language_editor_add (PrefsGeneralPage *general_page,
                      const char       *code,
                      const char       *desc)
 {
-  GtkWidget *event_box;
   GtkWidget *row;
-  GtkWidget *prefix;
-  GtkWidget *action;
+  GtkWidget *event_box;
   int len;
   int index;
 
   g_assert (code != NULL && desc != NULL);
 
-  len = get_list_box_length (general_page->lang_listbox);
+  len = get_list_box_length (GTK_LIST_BOX (general_page->lang_listbox));
 
-  for (index = 0; index < len; index++) {
+  for (index = 0; index < len - 1; index++) {
     GtkListBoxRow *widget;
-    char *row_code;
+    const char *row_code;
 
-    widget = gtk_list_box_get_row_at_index (GTK_LIST_BOX (general_page->lang_listbox), index++);
+    widget = gtk_list_box_get_row_at_index (GTK_LIST_BOX (general_page->lang_listbox), index);
 
-    row_code = g_object_get_data (G_OBJECT (widget), "code");
+    row_code = ephy_lang_row_get_code (EPHY_LANG_ROW (widget));
     if (row_code && strcmp (row_code, code) == 0)
       return;
   }
 
-  row = hdy_action_row_new ();
-  hdy_preferences_row_set_title (HDY_PREFERENCES_ROW (row), desc);
-  g_object_set_data (G_OBJECT (row), "code", g_strdup (code));
-  gtk_style_context_add_class (gtk_widget_get_style_context (GTK_WIDGET (row)), "draggable");
+  row = ephy_lang_row_new ();
 
-  event_box = gtk_event_box_new ();
-  gtk_drag_source_set (GTK_WIDGET (event_box), GDK_BUTTON1_MASK, entries, 1, GDK_ACTION_MOVE);
+  ephy_lang_row_set_code (EPHY_LANG_ROW (row), code);
+  ephy_lang_row_set_title (EPHY_LANG_ROW (row), desc);
+  gtk_style_context_add_class (gtk_widget_get_style_context (row), "row");
+
+  event_box = ephy_lang_row_get_drag_event_box (EPHY_LANG_ROW (row));
+  gtk_drag_source_set (event_box, GDK_BUTTON1_MASK, entries, 1, GDK_ACTION_MOVE);
   g_signal_connect (event_box, "drag-begin", G_CALLBACK (drag_begin), general_page);
   g_signal_connect (event_box, "drag-end", G_CALLBACK (drag_end), general_page);
   g_signal_connect (event_box, "drag-data-get", G_CALLBACK (drag_data_get), general_page);
-  prefix = gtk_image_new_from_icon_name ("list-drag-handle-symbolic", GTK_ICON_SIZE_SMALL_TOOLBAR);
-  gtk_container_add (GTK_CONTAINER (event_box), prefix);
-  hdy_action_row_add_prefix (HDY_ACTION_ROW (row), event_box);
 
-  action = gtk_button_new_from_icon_name ("edit-delete-symbolic", GTK_ICON_SIZE_SMALL_TOOLBAR);
-  gtk_widget_set_tooltip_text (action, _("Delete language"));
-  g_object_set_data (G_OBJECT (row), "action", action);
-  g_object_set_data (G_OBJECT (action), "row", row);
-  g_signal_connect (action, "clicked", G_CALLBACK (language_editor_remove_button_clicked_cb), general_page);
-  gtk_widget_set_valign (action, GTK_ALIGN_CENTER);
-  gtk_container_add (GTK_CONTAINER (row), action);
+  g_signal_connect (row, "delete-button-clicked", G_CALLBACK (language_editor_delete_button_clicked_cb), general_page);
 
-  gtk_widget_show_all (GTK_WIDGET (row));
-
-  gtk_list_box_insert (GTK_LIST_BOX (general_page->lang_listbox), GTK_WIDGET (row), len - 1);
+  gtk_list_box_insert (GTK_LIST_BOX (general_page->lang_listbox), row, len - 1);
 }
 
 static void
@@ -607,7 +556,7 @@ setup_add_language_dialog (PrefsGeneralPage *general_page)
   guint i, n;
   GtkBuilder *builder;
   GtkWidget *prefs_dialog;
-  g_auto (GStrv) locales;
+  g_auto (GStrv) locales = NULL;
 
   builder = gtk_builder_new_from_resource ("/org/gnome/epiphany/gtk/prefs-lang-dialog.ui");
   prefs_dialog = gtk_widget_get_toplevel (GTK_WIDGET (general_page));
@@ -744,43 +693,74 @@ add_system_language_entry (PrefsGeneralPage *general_page)
 }
 
 static void
-download_path_changed_cb (GtkFileChooser *button)
+download_folder_file_chooser_cb (GtkNativeDialog  *chooser,
+                                 GtkResponseType   response,
+                                 PrefsGeneralPage *general_page)
 {
-  char *dir;
+  if (response == GTK_RESPONSE_ACCEPT) {
+    g_autoptr (GFile) file = NULL;
+    g_autofree char *path = NULL;
 
-  dir = gtk_file_chooser_get_filename (button);
-  if (dir == NULL)
-    return;
+    file = gtk_file_chooser_get_file (GTK_FILE_CHOOSER (chooser));
+    path = g_file_get_path (file);
 
-  g_settings_set_string (EPHY_SETTINGS_STATE,
-                         EPHY_PREFS_STATE_DOWNLOAD_DIR, dir);
-  g_free (dir);
+    if (path)
+      g_settings_set_string (EPHY_SETTINGS_STATE,
+                             EPHY_PREFS_STATE_DOWNLOAD_DIR, path);
+  }
+
+  gtk_native_dialog_destroy (chooser);
 }
 
 static void
-create_download_path_button (PrefsGeneralPage *general_page)
+download_folder_row_activated_cb (PrefsGeneralPage *general_page)
 {
-  GtkWidget *button;
-  char *dir;
+  GtkWidget *parent;
+  g_autofree char *downloads_path = NULL;
+  GtkFileChooserNative *chooser;
 
-  dir = ephy_file_get_downloads_dir ();
+  parent = gtk_widget_get_toplevel (GTK_WIDGET (general_page));
+  chooser = gtk_file_chooser_native_new (_("Select a Directory"),
+                                         GTK_WINDOW (parent),
+                                         GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER,
+                                         _("_Select"),
+                                         _("_Cancel"));
+  gtk_native_dialog_set_modal (GTK_NATIVE_DIALOG (chooser), TRUE);
 
-  button = gtk_file_chooser_button_new (_("Select a directory"),
-                                        GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER);
+  downloads_path = ephy_file_get_downloads_dir ();
 
-  gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (button), dir);
-  gtk_file_chooser_button_set_width_chars (GTK_FILE_CHOOSER_BUTTON (button),
-                                           DOWNLOAD_BUTTON_WIDTH);
-  g_signal_connect (button, "selection-changed",
-                    G_CALLBACK (download_path_changed_cb), general_page);
-  gtk_container_add (GTK_CONTAINER (general_page->download_folder_row), button);
-  gtk_widget_set_valign (button, GTK_ALIGN_CENTER);
-  gtk_widget_show (button);
+  if (downloads_path && downloads_path[0]) {
+    g_autoptr (GFile) downloads_dir = NULL;
+    g_autoptr (GError) error = NULL;
 
-  g_settings_bind_writable (EPHY_SETTINGS_STATE,
-                            EPHY_PREFS_STATE_DOWNLOAD_DIR,
-                            button, "sensitive", FALSE);
-  g_free (dir);
+    downloads_dir = g_file_new_for_path (downloads_path);
+
+    gtk_file_chooser_set_current_folder_file (GTK_FILE_CHOOSER (chooser),
+                                              downloads_dir,
+                                              &error);
+
+    if (error)
+      g_warning ("Failed to set current folder %s: %s", downloads_path, error->message);
+  }
+
+  g_signal_connect (chooser, "response",
+                    G_CALLBACK (download_folder_file_chooser_cb),
+                    general_page);
+
+  gtk_native_dialog_show (GTK_NATIVE_DIALOG (chooser));
+}
+
+static gboolean
+download_folder_get_mapping (GValue   *value,
+                             GVariant *variant,
+                             gpointer  user_data)
+{
+  g_autofree char *path = ephy_file_get_downloads_dir ();
+  g_autoptr (GFile) dir = g_file_new_for_path (path);
+
+  g_value_take_string (value, ephy_file_get_display_name (dir));
+
+  return TRUE;
 }
 
 static gboolean
@@ -828,10 +808,10 @@ save_web_application (PrefsGeneralPage *general_page)
     changed = TRUE;
   }
 
-  text = (const char *)g_object_get_data (G_OBJECT (general_page->webapp_icon), "ephy-webapp-icon-url");
-  if (g_strcmp0 (general_page->webapp->icon_url, text) != 0) {
-    g_free (general_page->webapp->icon_url);
-    general_page->webapp->icon_url = g_strdup (text);
+  text = (const char *)g_object_get_data (G_OBJECT (general_page->webapp_icon), "ephy-webapp-icon-path");
+  if (g_strcmp0 (general_page->webapp->icon_path, text) != 0) {
+    g_free (general_page->webapp->icon_path);
+    general_page->webapp->icon_path = g_strdup (text);
     changed = TRUE;
   }
 
@@ -853,25 +833,20 @@ prefs_general_page_save_web_application (PrefsGeneralPage *general_page)
 
 static void
 prefs_general_page_update_webapp_icon (PrefsGeneralPage *general_page,
-                                       const char       *icon_url)
+                                       const char       *icon_path)
 {
-  GdkPixbuf *icon, *scaled_icon;
-  int icon_width, icon_height;
-  double scale;
+  g_autoptr (GdkPixbuf) icon = gdk_pixbuf_new_from_file (icon_path, NULL);
 
-  icon = gdk_pixbuf_new_from_file (icon_url, NULL);
   if (!icon)
     return;
 
-  icon_width = gdk_pixbuf_get_width (icon);
-  icon_height = gdk_pixbuf_get_height (icon);
-  scale = MIN ((double)64 / icon_width, (double)64 / icon_height);
-  scaled_icon = gdk_pixbuf_scale_simple (icon, icon_width * scale, icon_height * scale, GDK_INTERP_NEAREST);
-  g_object_unref (icon);
-  gtk_image_set_from_pixbuf (GTK_IMAGE (general_page->webapp_icon), scaled_icon);
-  g_object_set_data_full (G_OBJECT (general_page->webapp_icon), "ephy-webapp-icon-url",
-                          g_strdup (icon_url), g_free);
-  g_object_unref (scaled_icon);
+  gtk_image_set_from_gicon (GTK_IMAGE (general_page->webapp_icon),
+                            G_ICON (icon),
+                            GTK_ICON_SIZE_DND);
+  gtk_image_set_pixel_size (GTK_IMAGE (general_page->webapp_icon), 32);
+
+  g_object_set_data_full (G_OBJECT (general_page->webapp_icon), "ephy-webapp-icon-path",
+                          g_strdup (icon_path), g_free);
 }
 
 static void
@@ -880,11 +855,12 @@ webapp_icon_chooser_response_cb (GtkNativeDialog  *file_chooser,
                                  PrefsGeneralPage *general_page)
 {
   if (response == GTK_RESPONSE_ACCEPT) {
-    char *icon_url;
+    g_autoptr (GFile) icon_file = NULL;
+    g_autofree char *icon_path = NULL;
 
-    icon_url = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (file_chooser));
-    prefs_general_page_update_webapp_icon (general_page, icon_url);
-    g_free (icon_url);
+    icon_file = gtk_file_chooser_get_file (GTK_FILE_CHOOSER (file_chooser));
+    icon_path = g_file_get_path (icon_file);
+    prefs_general_page_update_webapp_icon (general_page, icon_path);
     prefs_general_page_save_web_application (general_page);
   }
 
@@ -1020,7 +996,7 @@ blank_homepage_get_mapping (GValue   *value,
   const char *setting;
 
   setting = g_variant_get_string (variant, NULL);
-  if (g_strcmp0 (setting, "about:blank") == 0)
+  if (g_strcmp0 (setting, "about:newtab") == 0)
     g_value_set_boolean (value, TRUE);
 
   return TRUE;
@@ -1038,7 +1014,7 @@ blank_homepage_set_mapping (const GValue       *value,
 
   gtk_entry_set_text (GTK_ENTRY (general_page->custom_homepage_entry), "");
 
-  return g_variant_new_string ("about:blank");
+  return g_variant_new_string ("about:newtab");
 }
 
 static gboolean
@@ -1049,7 +1025,7 @@ custom_homepage_get_mapping (GValue   *value,
   const char *setting;
 
   setting = g_variant_get_string (variant, NULL);
-  if (setting && setting[0] != '\0' && g_strcmp0 (setting, "about:blank") != 0)
+  if (setting && setting[0] != '\0' && g_strcmp0 (setting, "about:newtab") != 0)
     g_value_set_boolean (value, TRUE);
   return TRUE;
 }
@@ -1079,21 +1055,6 @@ custom_homepage_set_mapping (const GValue       *value,
   return g_variant_new_string (setting);
 }
 
-static void
-on_search_engine_dialog_button_clicked (GtkWidget        *button,
-                                        PrefsGeneralPage *general_page)
-{
-  GtkWindow *search_engine_dialog;
-  GtkWindow *prefs_dialog;
-
-  search_engine_dialog = GTK_WINDOW (ephy_search_engine_dialog_new ());
-  prefs_dialog = GTK_WINDOW (gtk_widget_get_toplevel (GTK_WIDGET (general_page)));
-
-  gtk_window_set_transient_for (search_engine_dialog, prefs_dialog);
-  gtk_window_set_modal (search_engine_dialog, TRUE);
-  gtk_window_present_with_time (search_engine_dialog, gtk_get_current_event_time ());
-}
-
 static void
 on_manage_webapp_additional_urls_button_clicked (GtkWidget        *button,
                                                  PrefsGeneralPage *general_page)
@@ -1106,7 +1067,7 @@ on_manage_webapp_additional_urls_button_clicked (GtkWidget        *button,
 
   gtk_window_set_transient_for (GTK_WINDOW (urls_dialog), prefs_dialog);
   gtk_window_set_modal (GTK_WINDOW (urls_dialog), TRUE);
-  gtk_window_present_with_time (GTK_WINDOW (urls_dialog), gtk_get_current_event_time ());
+  gtk_window_present (GTK_WINDOW (urls_dialog));
 }
 
 static void
@@ -1123,8 +1084,11 @@ prefs_general_page_class_init (PrefsGeneralPageClass *klass)
   /* Web Application */
   gtk_widget_class_bind_template_child (widget_class, PrefsGeneralPage, webapp_box);
   gtk_widget_class_bind_template_child (widget_class, PrefsGeneralPage, webapp_icon);
+  gtk_widget_class_bind_template_child (widget_class, PrefsGeneralPage, webapp_icon_row);
   gtk_widget_class_bind_template_child (widget_class, PrefsGeneralPage, webapp_url);
+  gtk_widget_class_bind_template_child (widget_class, PrefsGeneralPage, webapp_url_row);
   gtk_widget_class_bind_template_child (widget_class, PrefsGeneralPage, webapp_title);
+  gtk_widget_class_bind_template_child (widget_class, PrefsGeneralPage, webapp_title_row);
 
   /* Web Content */
   gtk_widget_class_bind_template_child (widget_class, PrefsGeneralPage, adblock_allow_switch);
@@ -1141,9 +1105,10 @@ prefs_general_page_class_init (PrefsGeneralPageClass *klass)
   gtk_widget_class_bind_template_child (widget_class, PrefsGeneralPage, download_box);
   gtk_widget_class_bind_template_child (widget_class, PrefsGeneralPage, ask_on_download_switch);
   gtk_widget_class_bind_template_child (widget_class, PrefsGeneralPage, download_folder_row);
+  gtk_widget_class_bind_template_child (widget_class, PrefsGeneralPage, download_folder_label);
 
   /* Search Engines */
-  gtk_widget_class_bind_template_child (widget_class, PrefsGeneralPage, search_box);
+  gtk_widget_class_bind_template_child (widget_class, PrefsGeneralPage, search_engine_group);
 
   /* Session */
   gtk_widget_class_bind_template_child (widget_class, PrefsGeneralPage, session_box);
@@ -1153,7 +1118,6 @@ prefs_general_page_class_init (PrefsGeneralPageClass *klass)
 
   /* Browsing */
   gtk_widget_class_bind_template_child (widget_class, PrefsGeneralPage, browsing_box);
-  gtk_widget_class_bind_template_child (widget_class, PrefsGeneralPage, enable_smooth_scrolling_switch);
   gtk_widget_class_bind_template_child (widget_class, PrefsGeneralPage, enable_mouse_gesture_switch);
   gtk_widget_class_bind_template_child (widget_class, PrefsGeneralPage, enable_switch_to_new_tab);
 
@@ -1166,7 +1130,7 @@ prefs_general_page_class_init (PrefsGeneralPageClass *klass)
   gtk_widget_class_bind_template_callback (widget_class, on_webapp_icon_button_clicked);
   gtk_widget_class_bind_template_callback (widget_class, on_webapp_entry_changed);
   gtk_widget_class_bind_template_callback (widget_class, on_manage_webapp_additional_urls_button_clicked);
-  gtk_widget_class_bind_template_callback (widget_class, on_search_engine_dialog_button_clicked);
+  gtk_widget_class_bind_template_callback (widget_class, download_folder_row_activated_cb);
 }
 
 static void
@@ -1178,7 +1142,6 @@ init_lang_listbox (PrefsGeneralPage *general_page)
   gtk_drag_dest_set (general_page->lang_listbox, GTK_DEST_DEFAULT_MOTION | GTK_DEST_DEFAULT_DROP, entries, 1, GDK_ACTION_MOVE);
   g_signal_connect (general_page->lang_listbox, "drag-data-received", G_CALLBACK (drag_data_received), general_page);
   g_signal_connect (general_page->lang_listbox, "drag-motion", G_CALLBACK (drag_motion), NULL);
-  g_signal_connect (general_page->lang_listbox, "drag-leave", G_CALLBACK (drag_leave), NULL);
 
   list = g_settings_get_strv (EPHY_SETTINGS_WEB,
                               EPHY_PREFS_WEB_LANGUAGE);
@@ -1211,12 +1174,17 @@ setup_general_page (PrefsGeneralPage *general_page)
   /* ======================================================================== */
   /* ========================== Web Application ============================= */
   /* ======================================================================== */
-  if (ephy_embed_shell_get_mode (ephy_embed_shell_get_default ()) == EPHY_EMBED_SHELL_MODE_APPLICATION) {
-    general_page->webapp = ephy_web_application_for_profile_directory (ephy_profile_dir ());
+  if (ephy_embed_shell_get_mode (ephy_embed_shell_get_default ()) == EPHY_EMBED_SHELL_MODE_APPLICATION &&
+      !ephy_is_running_inside_sandbox ()) {
+    general_page->webapp = ephy_web_application_for_profile_directory (ephy_profile_dir (),
+                                                                       EPHY_WEB_APP_NO_TMP_ICON);
     g_assert (general_page->webapp);
-    prefs_general_page_update_webapp_icon (general_page, general_page->webapp->icon_url);
-    gtk_entry_set_text (GTK_ENTRY (general_page->webapp_url), general_page->webapp->url);
-    gtk_entry_set_text (GTK_ENTRY (general_page->webapp_title), general_page->webapp->name);
+
+    if (!g_settings_get_boolean (EPHY_SETTINGS_WEB_APP, EPHY_PREFS_WEB_APP_SYSTEM)) {
+      prefs_general_page_update_webapp_icon (general_page, general_page->webapp->icon_path);
+      gtk_entry_set_text (GTK_ENTRY (general_page->webapp_url), general_page->webapp->url);
+      gtk_entry_set_text (GTK_ENTRY (general_page->webapp_title), general_page->webapp->name);
+    }
   }
 
   /* ======================================================================== */
@@ -1286,10 +1254,18 @@ setup_general_page (PrefsGeneralPage *general_page)
   /* ======================================================================== */
   /* ========================== Downloads =================================== */
   /* ======================================================================== */
-  if (ephy_is_running_inside_flatpak ())
+  if (ephy_is_running_inside_sandbox ())
     gtk_widget_hide (general_page->download_box);
   else
-    create_download_path_button (general_page);
+    g_settings_bind_with_mapping (EPHY_SETTINGS_STATE,
+                                  EPHY_PREFS_STATE_DOWNLOAD_DIR,
+                                  general_page->download_folder_label,
+                                  "label",
+                                  G_SETTINGS_BIND_GET,
+                                  download_folder_get_mapping,
+                                  NULL,
+                                  general_page,
+                                  NULL);
 
   g_settings_bind (web_settings,
                    EPHY_PREFS_WEB_ASK_ON_DOWNLOAD,
@@ -1324,12 +1300,6 @@ setup_general_page (PrefsGeneralPage *general_page)
   /* ======================================================================== */
   /* ========================== Browsing ==================================== */
   /* ======================================================================== */
-  g_settings_bind (web_settings,
-                   EPHY_PREFS_WEB_ENABLE_SMOOTH_SCROLLING,
-                   general_page->enable_smooth_scrolling_switch,
-                   "active",
-                   G_SETTINGS_BIND_DEFAULT);
-
   g_settings_bind (web_settings,
                    EPHY_PREFS_WEB_ENABLE_MOUSE_GESTURES,
                    general_page->enable_mouse_gesture_switch,
@@ -1359,18 +1329,25 @@ prefs_general_page_init (PrefsGeneralPage *general_page)
 {
   EphyEmbedShellMode mode = ephy_embed_shell_get_mode (ephy_embed_shell_get_default ());
 
+  g_type_ensure (EPHY_TYPE_SEARCH_ENGINE_LIST_BOX);
+
   gtk_widget_init_template (GTK_WIDGET (general_page));
 
+  setup_general_page (general_page);
+
   gtk_widget_set_visible (general_page->webapp_box,
-                          mode == EPHY_EMBED_SHELL_MODE_APPLICATION);
+                          mode == EPHY_EMBED_SHELL_MODE_APPLICATION &&
+                          !g_settings_get_boolean (EPHY_SETTINGS_WEB_APP,
+                                                   EPHY_PREFS_WEB_APP_SYSTEM));
+  gtk_widget_set_visible (general_page->webapp_icon_row, !ephy_is_running_inside_sandbox ());
+  gtk_widget_set_visible (general_page->webapp_url_row, !ephy_is_running_inside_sandbox ());
+  gtk_widget_set_visible (general_page->webapp_title_row, !ephy_is_running_inside_sandbox ());
   gtk_widget_set_visible (general_page->homepage_box,
                           mode != EPHY_EMBED_SHELL_MODE_APPLICATION);
-  gtk_widget_set_visible (general_page->search_box,
+  gtk_widget_set_visible (general_page->search_engine_group,
                           mode != EPHY_EMBED_SHELL_MODE_APPLICATION);
   gtk_widget_set_visible (general_page->session_box,
                           mode != EPHY_EMBED_SHELL_MODE_APPLICATION);
   gtk_widget_set_visible (general_page->browsing_box,
                           mode != EPHY_EMBED_SHELL_MODE_APPLICATION);
-
-  setup_general_page (general_page);
 }
